"""
Vector class for PETSc applications
"""

import copy
import numpy as np

from pymgrit.core.vector import Vector as PymgritVector

try:
    from petsc4py import PETSc
except ImportError as e:
    import sys

    sys.exit("This example requires petsc4py.")


class VectorPetsc(PymgritVector):
    """
    Vector class for PETSc vectors
    """

    def __init__(self, values: PETSc.Vec) -> None:
        """
        Constructor.

        :param values: PETSc.Vec with approximation
        """
        if isinstance(values, PETSc.Vec):
            self.values = copy.deepcopy(values)
        else:
            raise Exception('Wrong datatype')

    def __add__(self, other: '__class__') -> '__class__':
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        return VectorPetsc(self.get_values() + other.get_values())

    def __sub__(self, other: '__class__') -> '__class__':
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        return VectorPetsc(self.get_values() - other.get_values())

    def __mul__(self, other) -> '__class__':
        """
        Multiplication of a vector object and a float (self and other)

        :param other: object to be multiplied with self
        :return: difference of vector object self and input object other
        """
        return VectorPetsc(self.get_values() * other)

    def norm(self) -> float:
        """
        Norm of a vector object

        :return: Frobenius-norm of vector object
        """
        return self.values.norm(PETSc.NormType.FROBENIUS)

    def clone(self) -> '__class__':
        """
        Initialize vector object with copied values

        :rtype: vector object with zero values
        """

        return VectorPetsc(self.get_values())

    def clone_zero(self) -> '__class__':
        """
        Initialize vector object with zeros

        :rtype: vector object with zero values
        """

        return VectorPetsc(self.get_values() * 0)

    def clone_rand(self) -> '__class__':
        """
        Initialize vector object with random values

        :rtype: vector object with random values
        """
        # TODO
        return VectorPetsc(self.get_values())

    def set_values(self, values: PETSc.Vec) -> None:
        """
        Set vector data

        :param values: values for vector object
        """
        self.values = values

    def get_values(self) -> PETSc.Vec:
        """
        Get vector data

        :return: values of vector object
        """
        return self.values

    def pack(self) -> np.ndarray:
        """
        Pack data

        :return: values of vector object
        """
        return self.values.getArray()

    def unpack(self, values: np.ndarray) -> None:
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.values.setArray(values)



"""
Vector and application class for the 2D Allen-Cahn equation

Based on https://epubs.siam.org/doi/abs/10.1137/080738398?mobileUi=0 and
https://dl.acm.org/doi/10.1145/3310410
"""

import numpy as np
import scipy.sparse as sp
from scipy.sparse.linalg import spsolve

from pymgrit.core.vector import Vector
from pymgrit.core.application import Application


class VectorAllenCahn2D(Vector):
    """
    Vector class for the 2D Allen-Cahn equation
    """

    def __init__(self, nx, ny):
        """
        Constructor.

        :param nx: number of degrees of freedom in x-direction
        :param ny: number of degrees of freedom in y-direction
        """
        super().__init__()
        self.nx = nx
        self.ny = ny
        self.values = np.zeros((self.nx, self.ny))

    def __add__(self, other):
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        tmp = VectorAllenCahn2D(self.nx, self.ny)
        tmp.set_values(self.get_values() + other.get_values())
        return tmp

    def __sub__(self, other):
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        tmp = VectorAllenCahn2D(self.nx, self.ny)
        tmp.set_values(self.get_values() - other.get_values())
        return tmp

    def __mul__(self, other):
        """
        Multiplication of a vector object and a float (self and other)

        :param other: object to be multiplied with self
        :return: difference of vector object self and input object other
        """
        tmp = VectorAllenCahn2D(self.nx, self.ny)
        tmp.set_values(self.get_values() * other)
        return tmp

    def norm(self):
        """
        Norm of a vector object

        :return: 2-norm of vector object
        """
        return np.linalg.norm(self.values)

    def clone(self):
        """
        Clone vector object

        :rtype: vector object with zero values
        """
        tmp = VectorAllenCahn2D(self.nx, self.ny)
        tmp.set_values(self.get_values())
        return tmp

    def clone_zero(self):
        """
        Initialize vector object with zeros

        :rtype: vector object with zero values
        """
        return VectorAllenCahn2D(self.nx, self.ny)

    def clone_rand(self):
        """
        Initialize vector object with random values

        :rtype: vector object with random values
        """
        tmp = VectorAllenCahn2D(self.nx, self.ny)
        tmp.set_values(np.random.rand(self.nx, self.ny))
        return tmp

    def set_values(self, values):
        """
        Set vector data

        :param values: values for vector object
        """
        self.values = values

    def get_values(self):
        """
        Get vector data

        :return: values of vector object
        """
        return self.values

    def pack(self):
        """
        Pack data

        :return: values of vector object
        """
        return self.values

    def unpack(self, values):
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.values = values


class AllenCahn(Application):
    """
    Application class for the Allen-Cahn in 2D space,
        u_t =  (u_xx + u_yy) + 1/ eps^2 u(1-u),  a > 0,
        on [-0.5, 0.5] x [-0.5, 0.5] x (t_start, t_end],
    with periodic boundary conditions, eps > 0,
    and subject to the initial condition
        u(x,0)  = tanh((R0-|x|)/sqrt(2eps).
    """

    def __init__(self, nx=128, nu=2, eps=0.04, newton_maxiter=100, newton_tol=1e-12, lin_tol=1e-12, lin_maxiter=100,
                 radius=0.25, method='IMPL', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.nu = nu
        self.eps = eps
        self.newton_maxiter = newton_maxiter
        self.newton_tol = newton_tol
        self.lin_tol = lin_tol
        self.lin_maxiter = lin_maxiter
        self.radius = radius
        self.nx = nx
        self.ny = nx
        self.method = method

        if self.method == 'IMPL' or self.method == 'IMEX' or self.method == 'CN':
            pass
        else:
            raise Exception("Unknown method. Choose IMPL (implicit), IMEX (implicit-explicit) or CN (Crank-Nicolson")

        self.dx = 1.0 / self.nx
        self.space_disc = self.compute_matrix()
        self.id = sp.eye(self.nx * self.ny)
        self.x = np.linspace(start=-0.5, stop=0.5, num=self.nx)

        self.vector_t_start = self.initial_guess()
        self.vector_template = VectorAllenCahn2D(nx=self.nx, ny=self.ny)

    def compute_matrix(self):
        """
        Define spatial discretization matrix

        :return: space discretizaton
        """
        stencil = [1, -2, 1]
        zero_pos = 2
        dstencil = np.concatenate((stencil, np.delete(stencil, zero_pos - 1)))
        offsets = np.concatenate(([self.nx - i - 1 for i in reversed(range(zero_pos - 1))],
                                  [i - zero_pos + 1 for i in range(zero_pos - 1, len(stencil))]))
        doffsets = np.concatenate((offsets, np.delete(offsets, zero_pos - 1) - self.nx))

        space_disc = sp.diags(dstencil, doffsets, shape=(self.nx, self.ny), format='csc')
        space_disc = sp.kron(space_disc, sp.eye(self.nx)) + sp.kron(sp.eye(self.ny), space_disc)
        space_disc *= 1.0 / (self.dx ** 2)

        return space_disc

    def step(self, u_start: VectorAllenCahn2D, t_start: float, t_stop: float) -> VectorAllenCahn2D:
        """
        Time integration routine for 2D Allen-Cahn problem:

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        ret = VectorAllenCahn2D(self.nx, self.ny)
        if self.method == 'IMEX':
            new = u_start.get_values().flatten()
            rhs = new + (t_stop - t_start) * (1 / self.eps ** 2 * new * (1.0 - new ** self.nu))
            new = spsolve(self.id - (t_stop - t_start) * self.space_disc, rhs)
            ret.set_values(new.reshape(self.nx, self.ny))
        else:
            new = u_start.get_values().flatten()
            old = u_start.get_values().flatten()
            eps2 = self.eps ** 2

            if self.method == 'CN':
                fac = (t_stop - t_start) / 2
                rhs = old + (t_stop - t_start) / 2 * (
                        self.space_disc.dot(old) + 1.0 / eps2 * old * (1.0 - old ** self.nu))
            else:
                fac = (t_stop - t_start)
                rhs = old

            n = 0
            while 0 < self.newton_maxiter:
                g = new - fac * (self.space_disc.dot(new) + 1.0 / eps2 * new * (1.0 - new ** self.nu)) - rhs
                if np.linalg.norm(g, np.inf) < self.newton_tol:
                    break
                dg = self.id - fac * (
                        self.space_disc + 1.0 / eps2 * sp.diags((1.0 - (self.nu + 1) * new ** self.nu), offsets=0))
                new -= spsolve(dg, g)
                n += 1
            ret.set_values(new.reshape(self.nx, self.ny))
        return ret

    def initial_guess(self):
        """
        Set initial condition

        :return: Initial guess
        """
        initial = VectorAllenCahn2D(nx=self.nx, ny=self.ny)
        values = initial.get_values()
        for i in range(self.nx):
            for j in range(self.ny):
                r2 = self.x[i] ** 2 + self.x[j] ** 2
                values[i, j] = np.tanh((self.radius - np.sqrt(r2)) / (np.sqrt(2) * self.eps))
        initial.set_values(values=values)
        return initial

    def exact_radius(self, t):
        """
        Compute exact radius

        :return: Exact radius
        """
        return np.sqrt(max(self.radius ** 2 - 2.0 * t, 0))

    def compute_radius(self, u):
        """
        Compute radius of u

        :return: Radius
        """
        return np.sqrt(np.count_nonzero(u.get_values() >= 0.0) / np.pi) * self.dx



"""
Standard grid transfer operator.
Copies the spatial solution from one level to another.
"""

import copy

from pymgrit.core import grid_transfer
from pymgrit.core.vector import Vector


class GridTransferCopy(grid_transfer.GridTransfer):
    """
    Standard grid transfer.
    Copies the spatial solution from one level to another.
    This function is called for every time point.
    """

    def __init__(self):
        """
        Constructor.
        """
        super().__init__()

    def restriction(self, u: Vector) -> Vector:
        """
        Copies the spatial solution at one point in time
        of one level and restricts it to the same point in time
        on the next coarser level.

        :param u: Approximate solution at a time point of one level
        :return: Input vector u at the same time point on next coarser time grid
        :rtype: Vector
        """
        return u.clone()

    def interpolation(self, u: Vector) -> Vector:
        """
        Copies the spatial solution at one point in time
        of one level and restricts it to the same point in time
        on the next finer level.

        :param u: Approximate solution at a time point of one level
        :return: Input vector u at the same time point on next finer time grid
        :rtype: Vector
        """
        return u.clone()



"""
Vector and application class for the 1D heat equation
"""

import numpy as np
from scipy import sparse as sp
from scipy.sparse.linalg import spsolve
from scipy.sparse import identity

from pymgrit.core.application import Application
from pymgrit.core.vector import Vector


class VectorHeat1D(Vector):
    """
    Vector class for the 1D heat equation
    """

    def __init__(self, size):
        """
        Constructor.

        :param size: number of spatial degrees of freedom
        """
        super().__init__()
        self.size = size
        self.values = np.zeros(size)

    def __add__(self, other):
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        tmp = VectorHeat1D(self.size)
        tmp.set_values(self.get_values() + other.get_values())
        return tmp

    def __sub__(self, other):
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        tmp = VectorHeat1D(self.size)
        tmp.set_values(self.get_values() - other.get_values())
        return tmp

    def __mul__(self, other):
        """
        Multiplication of one vector object with a float (self and other)

        :param other: object to be multiplied with self
        :return: difference of vector object self and input object other
        """
        tmp = VectorHeat1D(self.size)
        tmp.set_values(self.get_values() * other)
        return tmp

    def norm(self):
        """
        Norm of a vector object

        :return: 2-norm of vector object
        """
        return np.linalg.norm(self.values)

    def clone(self):
        """
        Clone vector object

        :rtype: vector object with zero values
        """
        tmp = VectorHeat1D(self.size)
        tmp.set_values(self.get_values())
        return tmp

    def clone_zero(self):
        """
        Initialize vector object with zeros

        :rtype: vector object with zero values
        """
        return VectorHeat1D(self.size)

    def clone_rand(self):
        """
        Initialize vector object with random values

        :rtype: vector object with random values
        """
        tmp = VectorHeat1D(self.size)
        tmp.set_values(np.random.rand(self.size))
        return tmp

    def set_values(self, values):
        """
        Set vector data

        :param values: values for vector object
        """
        self.values = values

    def get_values(self):
        """
        Get vector data

        :return: values of vector object
        """
        return self.values

    def pack(self):
        """
        Pack data

        :return: values of vector object
        """
        return self.values

    def unpack(self, values):
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.values = values


class Heat1D(Application):
    """
    Application class for the heat equation in 1D space,
        u_t - a*u_xx = b(x,t),  a > 0, x in [x_start,x_end], t in [0,T],
    with homogeneous Dirichlet boundary conditions in space.
    """

    def __init__(self, x_start, x_end, nx, a, init_cond=lambda x: x * 0, rhs=lambda x, t: x * 0, *args, **kwargs):
        """
        Constructor.

        :param x_start: left interval bound of spatial domain
        :param x_end: right interval bound of spatial domain
        :param nx: number of spatial degrees of freedom
        :param a: thermal conductivity
        :param init_cond: initial condition
        :param rhs: right-hand side
        """

        super().__init__(*args, **kwargs)
        # Spatial domain with homogeneous Dirichlet boundary conditions
        self.x_start = x_start
        self.x_end = x_end
        self.x = np.linspace(self.x_start, self.x_end, nx)
        self.x = self.x[1:-1]
        self.nx = nx - 2
        self.dx = self.x[1] - self.x[0]

        # Thermal conductivity
        self.a = a

        # Set (spatial) identity matrix and spatial discretization matrix
        self.identity = identity(self.nx, dtype='float', format='csr')
        self.space_disc = self.compute_matrix()

        # Set right-hand side routine
        self.rhs = rhs

        # Set the data structure for any user-defined time point
        self.vector_template = VectorHeat1D(self.nx)

        # Set initial condition
        self.init_cond = init_cond
        self.vector_t_start = VectorHeat1D(self.nx)
        self.vector_t_start.set_values(self.init_cond(self.x))

    def compute_matrix(self):
        """
        Define spatial discretization matrix for 1D heat equation

        Second-order central finite differences with matrix stencil
           (a / dx^2) * [-1  2  -1]
        """

        fac = self.a / self.dx ** 2

        diagonal = np.ones(self.nx) * 2 * fac
        lower = np.ones(self.nx - 1) * -fac
        upper = np.ones(self.nx - 1) * -fac

        matrix = sp.diags(
            diagonals=[diagonal, lower, upper],
            offsets=[0, -1, 1], shape=(self.nx, self.nx),
            format='csr')

        return matrix

    def step(self, u_start: VectorHeat1D, t_start: float, t_stop: float) -> VectorHeat1D:
        """
        Time integration routine for 1D heat equation example problem:
            Backward Euler (BDF1)

        One-step method
           u_i = (I + dt*L)^{-1} * (u_{i-1} + dt*b_i),
        where L = self.space_disc is the spatial discretization operator

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        tmp = u_start.get_values()
        tmp = spsolve((t_stop - t_start) * self.space_disc + self.identity,
                      tmp + self.rhs(self.x, t_stop) * (t_stop - t_start))
        ret = VectorHeat1D(len(tmp))
        ret.set_values(tmp)
        return ret



"""
MGRIT optimized for the induction
machine model "im_3kW". (https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines)
"""

import numpy as np

from pymgrit.core.mgrit import Mgrit


class MgritMachine(Mgrit):
    """
    MGRIT optimized for the getdp induction machine
    """

    def __init__(self, *args, **kwargs) -> None:
        """
        MGRIT optimized for the getdp induction machine
        """
        super().__init__(*args, **kwargs)

    def nested_iteration(self) -> None:
        """
        Generates an initial approximation on the finest grid
        by solving the problem on the coarsest grid and interpolating
        the approximation to the finest level.

        Performs nested iterations with a sin rhs
        """
        change = False
        tmp_problem_pwm = np.zeros(len(self.problem))
        if self.problem[0].pwm:
            change = True
            for lvl in range(len(self.problem)):
                tmp_problem_pwm[lvl] = self.problem[lvl].pwm
                self.problem[lvl].fopt[-1] = 0

        self.forward_solve(self.lvl_max - 1)

        for lvl in range(self.lvl_max - 2, -1, -1):
            for i in range(len(self.index_local[lvl + 1])):
                self.u[lvl][self.index_local_c[lvl][i]] = self.interpolation[lvl](
                    u=self.u[lvl + 1][self.index_local[lvl + 1][i]])

            self.f_exchange(lvl)
            self.c_exchange(lvl)
            if lvl > 0:
                self.iteration(lvl, 'V', 0, True)

        if change:
            for lvl in range(len(self.problem)):
                self.problem[lvl].fopt[-1] = tmp_problem_pwm[lvl]



"""
Abstract grid transfer class for user-defined grid transfer classes.
Every user-defined grid transfer class must inherit from this class.
Allows for additional spatial coarsening and refinement between time levels.

Required functions:
  - restriction
  - interpolation
"""
from abc import ABC, abstractmethod

from pymgrit.core.vector import Vector


class GridTransfer(ABC):
    """
    Abstract grid transfer class for user-defined grid transfer classes.
    Every user-defined grid transfer class must inherit from this class.
    Allows for additional spatial coarsening and refinement between time levels.

    Required functions:
      - restriction
      - interpolation
    """

    def __init__(self):
        """
        Constructor
        """

    @abstractmethod
    def restriction(self, u: Vector) -> Vector:
        """
        Spatial restriction.
        Receives the spatial solution at a point in time
        of one level and restricts it to the same point in time
        on the next coarser level.

        :param u: Approximate solution at a time point of one level
        :return: Restricted input vector u on next coarser time grid
        :rtype: Vector
        """

    @abstractmethod
    def interpolation(self, u: Vector) -> Vector:
        """
        Spatial interpolation.
        Receives the spatial solution at a point in time
        of one level and interpolates it to the same point in
        time on the next finer level.

        :param u: Approximate solution at a time point of one level
        :return: Interpolated input vector u on next finer time grid
        :rtype: Vector
        """



"""
Application class for 1D heat problem using BDF2 time integration

Note: values at two consecutive time points are grouped as pairs
"""

from typing import Callable
import numpy as np
from scipy import sparse as sp
from scipy.sparse.linalg import spsolve
from scipy.sparse import identity

from pymgrit.core.application import Application
from pymgrit.heat.vector_heat_1d_2pts import VectorHeat1D2Pts


class Heat1DBDF2(Application):
    """
    Application class for the heat equation in 1D space,
        u_t - a*u_xx = b(x,t),  a > 0, x in [x_start, x_end], t in [t_start, t_end],
    with homogeneous Dirichlet boundary conditions in space
    """

    def __init__(self, x_start: float, x_end: float, nx: int, dtau: float, a: float,
                 init_cond: Callable = lambda x: x * 0, rhs: Callable = lambda x, t: x * 0, *args,
                 **kwargs):
        """
        Constructor.

        :param x_start: left interval bound of spatial domain
        :param x_end: right interval bound of spatial domain
        :param nx: number of spatial degrees of freedom
        :param dtau: time-step size within pair
        :param a: thermal conductivity
        :param init_cond: initial condition
        :param rhs: right-hand side
        """

        super().__init__(*args, **kwargs)
        # Spatial domain with homogeneous Dirichlet boundary conditions
        self.x_start = x_start
        self.x_end = x_end
        self.x = np.linspace(self.x_start, self.x_end, nx)
        self.x = self.x[1:-1]
        self.nx = nx - 2
        self.dx = self.x[1] - self.x[0]

        # Thermal conductivity
        self.a = a

        # (Spatial) identity matrix and spatial discretization matrix
        self.identity = identity(self.nx, dtype='float', format='csr')
        self.space_disc = self.compute_matrix()

        # Set right-hand side routine
        self.rhs = rhs

        # Set the data structure for any user-defined time-point pairs
        self.vector_template = VectorHeat1D2Pts(self.nx, dtau)

        # Set initial condition
        self.init_cond = init_cond
        self.vector_t_start = VectorHeat1D2Pts(self.nx, dtau)
        tmp1 = self.init_cond(self.x)
        # Use trapezoidal rule to get value at time dtau
        tmp2 = spsolve((dtau / 2) * self.space_disc + self.identity,
                       (self.identity - (dtau / 2) * self.space_disc) * tmp1 +
                       (dtau / 2) * (self.rhs(self.x, self.t[0]) + self.rhs(self.x, self.t[0] + dtau)))
        self.vector_t_start.set_values(first_time_point=tmp1, second_time_point=tmp2, dtau=dtau)

    def compute_matrix(self):
        """
        Define spatial discretization matrix for heat equation problem.

        Discretization is centered finite differences with matrix stencil
           (a / dx^2) * [-1  2  -1]
        """

        fac = self.a / self.dx ** 2

        diagonal = np.ones(self.nx) * 2 * fac
        lower = np.ones(self.nx - 1) * -fac
        upper = np.ones(self.nx - 1) * -fac

        matrix = sp.diags(
            diagonals=[diagonal, lower, upper],
            offsets=[0, -1, 1], shape=(self.nx, self.nx),
            format='csr')

        return matrix

    def step(self, u_start: VectorHeat1D2Pts, t_start: float, t_stop: float) -> VectorHeat1D2Pts:
        """
        Time integration routine for 1D heat equation:
            BDF2

        Two-step method on variably spaced grid with spacing tau_i = t_i - t_{i-1}.
        In time-based stencil notation, we have at time point t_i
           [r_i^2/(tau_i*(1+r_i))*I,  -((1+r_i)/tau_i)*I,  (1+2r_i)/(tau_i*(1+r_i))*I + L,  0,  0],
        where L = self.space_disc is the spatial discretization operator and r_i = tau_i/tau_{i-1}

        Note: For the pair associated with input time t_stop
          * update at t_stop involves values at t_start and (t_start + dtau)
          * update at t_stop + dtau involves values at (t_start + dtau) and t_stop

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution for the input time t_stop
        """
        first, second, dtau = u_start.get_values()

        # Update value at t_i = t_stop
        tau_i = t_stop - t_start - dtau
        tau_im1 = dtau
        r_i = tau_i / tau_im1
        coeffm2 = (r_i ** 2) / (tau_i * (1 + r_i))
        coeffm1 = (1 + r_i) / tau_i
        coeff = (1 + 2 * r_i) / (tau_i * (1 + r_i))
        rhs = self.rhs(self.x, t_stop) - coeffm2 * first + coeffm1 * second

        tmp1 = spsolve(self.space_disc + coeff * self.identity, rhs)

        # Update value at t_i = t_stop + dtau
        tau_im1 = tau_i
        tau_i = dtau
        r_i = tau_i / tau_im1
        coeffm2 = (r_i ** 2) / (tau_i * (1 + r_i))
        coeffm1 = (1 + r_i) / tau_i
        coeff = (1 + 2 * r_i) / (tau_i * (1 + r_i))
        rhs = self.rhs(self.x, t_stop + dtau) - coeffm2 * second + coeffm1 * tmp1

        tmp2 = spsolve(self.space_disc + coeff * self.identity, rhs)

        ret = VectorHeat1D2Pts(u_start.size, u_start.dtau)
        ret.set_values(first_time_point=tmp1, second_time_point=tmp2, dtau=dtau)

        return ret



from .advection.advection_1d import Advection1D
from .arenstorf_orbit.arenstorf_orbit import ArenstorfOrbit
from .core.application import Application
from .brusselator.brusselator import Brusselator
from .core.grid_transfer import GridTransfer
from .core.grid_transfer_copy import GridTransferCopy
from .core.mgrit import Mgrit
from .core.simple_setup_problem import simple_setup_problem
from .core.vector import Vector
from .dahlquist.dahlquist import Dahlquist
from .heat.heat_1d import Heat1D
from .heat.heat_2d import Heat2D
from .heat.heat_1d_2pts_bdf1 import Heat1DBDF1
from .heat.heat_1d_2pts_bdf2 import Heat1DBDF2
from .heat.vector_heat_1d_2pts import VectorHeat1D2Pts

__all__ = [s for s in dir() if not s.startswith('_')]



"""
Vector and application class for the Arenstorf orbit problem
"""

import numpy as np
import matplotlib.pyplot as plt

from scipy.integrate import solve_ivp

from pymgrit.core.application import Application
from pymgrit.core.vector import Vector


class VectorArenstorfOrbit(Vector):
    """
    Vector for the Arenstorf orbit problem
    """

    def __init__(self):
        super().__init__()
        self.y0 = 0
        self.y1 = 0
        self.y2 = 0
        self.y3 = 0

    def __add__(self, other):
        tmp = VectorArenstorfOrbit()
        tmp.set_values(self.get_values() + other.get_values())
        return tmp

    def __sub__(self, other):
        tmp = VectorArenstorfOrbit()
        tmp.set_values(self.get_values() - other.get_values())
        return tmp

    def __mul__(self, other):
        tmp = VectorArenstorfOrbit()
        tmp.set_values(self.get_values() * other)
        return tmp

    def norm(self):
        return np.linalg.norm(np.array([self.y0, self.y1, self.y2, self.y3]))

    def clone(self):
        tmp = VectorArenstorfOrbit()
        tmp.set_values(self.get_values())
        return tmp

    def clone_zero(self):
        return VectorArenstorfOrbit()

    def clone_rand(self):
        tmp = VectorArenstorfOrbit()
        tmp.set_values(np.random.rand(4))
        return tmp

    def set_values(self, values):
        self.y0 = values[0]
        self.y1 = values[1]
        self.y2 = values[2]
        self.y3 = values[3]

    def get_values(self):
        return np.array([self.y0, self.y1, self.y2, self.y3])

    def pack(self):
        return np.array([self.y0, self.y1, self.y2, self.y3])

    def unpack(self, values):
        self.y0 = values[0]
        self.y1 = values[1]
        self.y2 = values[2]
        self.y3 = values[3]

    def plot(self):
        plt.plot(self.y0, self.y1, color='red', marker='.')


def arenstorf(t, y, a, b):
    d1 = ((y[0] + a) ** 2 + y[1] ** 2) ** (3 / 2)
    d2 = ((y[0] - b) ** 2 + y[1] ** 2) ** (3 / 2)
    yp = np.array([
        y[2],
        y[3],
        y[0] + 2 * y[3] - b * (y[0] + a) / d1 - a * (y[0] - b) / d2,
        y[1] - 2 * y[2] - b * y[1] / d1 - a * y[1] / d2
    ], dtype=float)
    return yp


class ArenstorfOrbit(Application):
    """
    Application for Arenstorf orbit problem,
       x'' = x + 2y' - b*(x + a)/D_1 - a*(x - b)/D_2,
       y'' = y - 2x' - b*y/D_1 - a*y/D_2
    with a = 0.012277471, b = 1 - a,
        D_1 = ((x + a)^2 + y^2)^(3/2),  D_2 = ((x - a)^2 + y^2)^(3/2)
    and ICs
       x(0) = 0.994,  x'(0) = 0,  y(0) = 0,  y'(0) = -2.00158510637908
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.vector_template = VectorArenstorfOrbit()  # Setting the class which is used for each time point
        self.vector_t_start = VectorArenstorfOrbit()
        self.vector_t_start.set_values(np.array([0.994, 0.0, 0.0, -2.00158510637908]))

        self.a = 0.012277471
        self.b = 1 - self.a

    def step(self, u_start: VectorArenstorfOrbit, t_start: float, t_stop: float) -> VectorArenstorfOrbit:
        res = solve_ivp(fun=arenstorf, y0=u_start.get_values(), args=(self.a, self.b),
                        t_span=np.array([t_start, t_stop]),
                        t_eval=np.array([t_start, t_stop]), method='RK45')
        ret = VectorArenstorfOrbit()
        ret.set_values(res.y[:, -1])
        return ret



"""
Application classes for Burgers equation
"""

try:
    from firedrake import FunctionSpace, Constant, TestFunction, TrialFunction, Function, FacetNormal, inner, dx, grad
    from firedrake import outer, LinearVariationalProblem, NonlinearVariationalSolver, dS, exp, SpatialCoordinate, avg
    from firedrake import sin, NonlinearVariationalProblem, pi, VectorFunctionSpace, project, as_vector, dot, nabla_grad
except ImportError as e:
    import sys

    sys.exit("This example requires firedrake. See https://pymgrit.github.io/pymgrit/coupling/firedrake.html")

from mpi4py import MPI

from pymgrit.core.application import Application
from pymgrit.firedrake.vector_firedrake import VectorFiredrake


class Burgers1D(Application):
    """
    Application class for the 1d Burgers equation
    """

    def __init__(self, mesh: object, nu: float, comm_space: MPI.Comm, *args, **kwargs):
        """
        Constructor
        """
        super().__init__(*args, **kwargs)

        # Spatial domain and function space
        self.mesh = mesh
        self.nu = nu
        self.comm_space = comm_space

        self.function_space = FunctionSpace(mesh, "Lagrange", 2)

        # Placeholder for time step - will be updated in the update method
        self.dt = Constant(0.)

        # Things we need for the form
        self.u_n1 = Function(self.function_space)
        self.u_n = Function(self.function_space)
        v = TestFunction(self.function_space)

        f = (((self.u_n1 - self.u_n) / self.dt) * v +
             self.u_n1 * self.u_n1.dx(0) * v +
             nu * self.u_n1.dx(0) * v.dx(0)) * dx

        # Setup problem and solver
        problem = NonlinearVariationalProblem(f, self.u_n1)
        self.solver = NonlinearVariationalSolver(problem, solver_parameters={"ksp_type": "preonly",
                                                                             "pc_type": "lu"})

        # Set the data structure for any user-defined time point
        self.vector_template = VectorFiredrake(self.u_n1)

        # Set initial condition:
        self.u_n.interpolate(sin(2 * pi * SpatialCoordinate(mesh)[0]))
        self.vector_t_start = VectorFiredrake(self.u_n)

    def step(self, u_start: VectorFiredrake, t_start: float, t_stop: float) -> VectorFiredrake:
        """
        Time integration routine for 1D Burgers problem:
            Backward Euler

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        self.dt.assign(t_stop - t_start)
        self.u_n.assign(u_start.get_values())
        self.solver.solve()
        return VectorFiredrake(self.u_n1)


class Burgers2D(Application):
    """
    Application class for the 2d Burgers equation
    """

    def __init__(self, mesh: object, nu: float, comm_space: MPI.Comm, *args, **kwargs):
        """
        Constructor
        """
        super().__init__(*args, **kwargs)

        # Spatial domain and function space
        self.mesh = mesh
        self.nu = nu
        self.comm_space = comm_space

        self.function_space = VectorFunctionSpace(mesh, "CG", 2)

        # Placeholder for time step - will be updated in the update method
        self.dt = Constant(0.)

        # Things we need for the form
        self.u_n1 = Function(self.function_space)
        self.u_n = Function(self.function_space)
        v = TestFunction(self.function_space)

        f = (inner((self.u_n1 - self.u_n) / self.dt, v) + inner(dot(self.u_n1, nabla_grad(self.u_n1)), v) + nu * inner(
            grad(self.u_n1), grad(v))) * dx

        # Setup problem and solver
        problem = NonlinearVariationalProblem(f, self.u_n1)
        self.solver = NonlinearVariationalSolver(problem, solver_parameters={"ksp_type": "preonly",
                                                                             "pc_type": "lu"})

        # Set the data structure for any user-defined time point
        self.vector_template = VectorFiredrake(self.u_n1)

        # Set initial condition:
        self.u_n.assign(project(as_vector([sin(pi * SpatialCoordinate(mesh)[0]), 0]), self.function_space))
        self.vector_t_start = VectorFiredrake(self.u_n)

    def step(self, u_start: VectorFiredrake, t_start: float, t_stop: float) -> VectorFiredrake:
        """
        Time integration routine for 2D burgers problem:
            Backward Euler

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        self.dt.assign(t_stop - t_start)
        self.u_n.assign(u_start.get_values())
        self.solver.solve()

        return VectorFiredrake(self.u_n1)



"""
Application class for the 2D Gray-Scott equation

Implementation follows:
https://github.com/Parallel-in-Time/pySDC/blob/master/pySDC/implementations/problem_classes/GrayScott_2D_PETSc_periodic.py
"""

import numpy as np

from mpi4py import MPI

from pymgrit.petsc.vector_petsc import VectorPetsc
from pymgrit.core.application import Application

try:
    from petsc4py import PETSc
except ImportError as e:
    import sys

    sys.exit("This example requires petsc4py.")


class GrayScott(Application):
    """
    Application class implementing the 2D Gray-Scott reaction-diffusion equation with periodic BC
    """

    def __init__(self, dmda: PETSc.DMDA, comm_x: MPI.Comm, du=1.0, dv=0.01, a=0.09, b=0.086, lsol_tol: float = 1e-10,
                 nlsol_tol: float = 1e-10, lsol_maxiter: int = 100, nlsol_maxiter: int = 100, L=1.0,
                 method='IMPL', *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.dmda = dmda
        self.mx, self.my = self.dmda.sizes
        self.L = L
        self.dx = self.L / self.mx
        self.dy = self.L / self.my
        (self.xs, self.xe), (self.ys, self.ye) = self.dmda.getRanges()
        self.du = du
        self.dv = dv
        self.a = a
        self.b = b
        self.method = method
        self.local_x = dmda.createLocalVec()

        if self.method == 'IMPL' or self.method == 'IMEX' or self.method == 'EXPL':
            pass
        else:
            raise Exception("Unknown method. Choose IMPL (implicit) or IMEX (implicit-explicit)")

        self.space_disc = self.compute_matrix()
        self.id = self.get_id()

        self.vector_template = VectorPetsc(self.dmda.createGlobalVec())
        self.vector_t_start = self.initial_guess()

        if self.method == 'IMEX':
            self.ksp = PETSc.KSP()
            self.ksp.create(comm=comm_x)
            self.ksp.setType('cg')
            pc = self.ksp.getPC()
            pc.setType('none')
            self.ksp.setInitialGuessNonzero(True)
            self.ksp.setFromOptions()
            self.ksp.setTolerances(rtol=lsol_tol, atol=lsol_tol, max_it=lsol_maxiter)
            self.dt = self.t[1] - self.t[0]
            self.ksp.setOperators(self.id - self.dt * self.space_disc)
        elif self.method == 'IMPL':
            self.snes = PETSc.SNES()
            self.snes.create(comm=comm_x)
            self.snes.setFromOptions()
            self.snes.setTolerances(rtol=nlsol_tol, atol=nlsol_tol, stol=nlsol_tol,
                                    max_it=nlsol_maxiter)

    def step(self, u_start: VectorPetsc, t_start: float, t_stop: float) -> VectorPetsc:
        """
        Time integration routine for 2D Gray Scott example problem

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        result = self.dmda.createGlobalVec()
        if self.method == 'IMEX':
            if not np.isclose(self.dt, (t_stop - t_start)):
                self.ksp.setOperators(self.id - (t_stop - t_start) * self.space_disc)
                self.dt = (t_stop - t_start)
            rhs = self.compute_rhs_imex(u_start=u_start, t_start=t_start, t_stop=t_stop)
            self.ksp.solve(rhs, result)
        elif self.method == 'IMPL':
            self.local_x = self.dmda.createLocalVec()
            self.dt = (t_stop - t_start)
            F = self.dmda.createGlobalVec()
            self.snes.setFunction(self.formFunction, F)
            J = self.dmda.createMatrix()
            self.snes.setJacobian(self.formJacobian, J)
            self.snes.solve(u_start.get_values(), result)
        else:
            f = self.dmda.createGlobalVec()
            u = u_start.get_values()

            fa = self.dmda.getVecArray(f)
            xa = self.dmda.getVecArray(u)

            self.space_disc.mult(u, f)
            self.diff_A = xa[:, :, 0] + (t_stop - t_start) * (
                    fa[:, :, 0] - xa[:, :, 0] * xa[:, :, 1] ** 2 + self.a * (1 - xa[:, :, 0]))
            self.diff_B = xa[:, :, 1] + (t_stop - t_start) * (
                    fa[:, :, 1] + xa[:, :, 0] * xa[:, :, 1] ** 2 - self.b * xa[:, :, 1])

            fa = self.dmda.getVecArray(result)
            fa[:, :, 0] = self.diff_A
            fa[:, :, 1] = self.diff_B
        return VectorPetsc(result)

    def compute_matrix(self) -> PETSc.Mat:
        """
        Helper function to assemble PETSc matrix A

        Returns:
            PETSc matrix object
        """
        space_disc = self.dmda.createMatrix()
        space_disc.setType('aij')
        space_disc.setFromOptions()
        space_disc.setPreallocationNNZ((5, 5))
        space_disc.setUp()

        space_disc.zeroEntries()
        row = PETSc.Mat.Stencil()
        col = PETSc.Mat.Stencil()
        for j in range(self.ys, self.ye):
            for i in range(self.xs, self.xe):
                row.index = (i, j)
                row.field = 0
                space_disc.setValueStencil(row, row, self.du * (-2.0 / self.dx ** 2 - 2.0 / self.dy ** 2))
                row.field = 1
                space_disc.setValueStencil(row, row, self.dv * (-2.0 / self.dx ** 2 - 2.0 / self.dy ** 2))
                # if j > 0:
                col.index = (i, j - 1)
                col.field = 0
                row.field = 0
                space_disc.setValueStencil(row, col, self.du / self.dy ** 2)
                col.field = 1
                row.field = 1
                space_disc.setValueStencil(row, col, self.dv / self.dy ** 2)
                # if j < my - 1:
                col.index = (i, j + 1)
                col.field = 0
                row.field = 0
                space_disc.setValueStencil(row, col, self.du / self.dy ** 2)
                col.field = 1
                row.field = 1
                space_disc.setValueStencil(row, col, self.dv / self.dy ** 2)
                # if i > 0:
                col.index = (i - 1, j)
                col.field = 0
                row.field = 0
                space_disc.setValueStencil(row, col, self.du / self.dx ** 2)
                col.field = 1
                row.field = 1
                space_disc.setValueStencil(row, col, self.dv / self.dx ** 2)
                # if i < mx - 1:
                col.index = (i + 1, j)
                col.field = 0
                row.field = 0
                space_disc.setValueStencil(row, col, self.du / self.dx ** 2)
                col.field = 1
                row.field = 1
                space_disc.setValueStencil(row, col, self.dv / self.dx ** 2)
        space_disc.assemble()

        return space_disc

    def get_id(self) -> PETSc.Mat:
        """
        Define identity matrix
        :return: identity matrix
        """
        id = self.dmda.createMatrix()
        id.setType('aij')
        id.setFromOptions()
        id.setPreallocationNNZ((1, 1))
        id.setUp()

        id.zeroEntries()
        row = PETSc.Mat.Stencil()
        for j in range(self.ys, self.ye):
            for i in range(self.xs, self.xe):
                for indx in [0, 1]:
                    row.index = (i, j)
                    row.field = indx
                    id.setValueStencil(row, row, 1.0)

        id.assemble()
        return id

    def compute_rhs_imex(self, u_start: VectorPetsc, t_start: float, t_stop: float) -> PETSc.Vec:
        """
        Right-hand side of spatial system

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: right-hand side of spatial system at each time step in case of implicit time integration
        """
        f = self.dmda.createGlobalVec()
        u = u_start.get_values()

        fa = self.dmda.getVecArray(f)
        xa = self.dmda.getVecArray(u)

        fa[:, :, 0] = xa[:, :, 0] + (t_stop - t_start) * (-xa[:, :, 0] * xa[:, :, 1] ** 2 + self.a * (1 - xa[:, :, 0]))
        fa[:, :, 1] = xa[:, :, 1] + (t_stop - t_start) * (xa[:, :, 0] * xa[:, :, 1] ** 2 - self.b * xa[:, :, 1])

        return f

    def formFunction(self, snes, X, F):
        """
        Function to evaluate the residual for the Newton solver
        This function should be equal to the RHS in the solution
        """
        self.dmda.globalToLocal(X, self.local_x)
        x = self.dmda.getVecArray(self.local_x)
        f = self.dmda.getVecArray(F)
        for j in range(self.ys, self.ye):
            for i in range(self.xs, self.xe):
                u = x[i, j]  # center
                u_e = x[i + 1, j]  # east
                u_w = x[i - 1, j]  # west
                u_s = x[i, j + 1]  # south
                u_n = x[i, j - 1]  # north
                u_xx = (u_e - 2 * u + u_w)
                u_yy = (u_n - 2 * u + u_s)
                f[i, j, 0] = x[i, j, 0] - \
                             (self.dt * (self.du * (u_xx[0] / self.dx ** 2 + u_yy[0] / self.dy ** 2) -
                                         x[i, j, 0] * x[i, j, 1] ** 2 + self.a * (1 - x[i, j, 0])))
                f[i, j, 1] = x[i, j, 1] - \
                             (self.dt * (self.dv * (u_xx[1] / self.dx ** 2 + u_yy[1] / self.dy ** 2) +
                                         x[i, j, 0] * x[i, j, 1] ** 2 - self.b * x[i, j, 1]))

    def formJacobian(self, snes, X, J, P):
        """
        Function to return the Jacobian matrix
        """
        self.dmda.globalToLocal(X, self.local_x)
        x = self.dmda.getVecArray(self.local_x)
        P.zeroEntries()
        row = PETSc.Mat.Stencil()
        col = PETSc.Mat.Stencil()

        for j in range(self.ys, self.ye):
            for i in range(self.xs, self.xe):
                # diagnoal 2-by-2 block (for u and v)
                row.index = (i, j)
                col.index = (i, j)
                row.field = 0
                col.field = 0
                val = 1.0 - self.dt * (self.du * (-2.0 / self.dx ** 2 - 2.0 / self.dy ** 2) -
                                       x[i, j, 1] ** 2 - self.a)
                P.setValueStencil(row, col, val)
                row.field = 0
                col.field = 1
                val = self.dt * 2.0 * x[i, j, 0] * x[i, j, 1]
                P.setValueStencil(row, col, val)
                row.field = 1
                col.field = 1
                val = 1.0 - self.dt * (self.dv * (-2.0 / self.dx ** 2 - 2.0 / self.dy ** 2) +
                                       2.0 * x[i, j, 0] * x[i, j, 1] - self.b)
                P.setValueStencil(row, col, val)
                row.field = 1
                col.field = 0
                val = -self.dt * x[i, j, 1] ** 2
                P.setValueStencil(row, col, val)

                # coupling through finite difference part
                col.index = (i, j - 1)
                col.field = 0
                row.field = 0
                P.setValueStencil(row, col, -self.dt * self.du / self.dx ** 2)
                col.field = 1
                row.field = 1
                P.setValueStencil(row, col, -self.dt * self.dv / self.dy ** 2)
                col.index = (i, j + 1)
                col.field = 0
                row.field = 0
                P.setValueStencil(row, col, -self.dt * self.du / self.dx ** 2)
                col.field = 1
                row.field = 1
                P.setValueStencil(row, col, -self.dt * self.dv / self.dy ** 2)
                col.index = (i - 1, j)
                col.field = 0
                row.field = 0
                P.setValueStencil(row, col, -self.dt * self.du / self.dx ** 2)
                col.field = 1
                row.field = 1
                P.setValueStencil(row, col, -self.dt * self.dv / self.dy ** 2)
                col.index = (i + 1, j)
                col.field = 0
                row.field = 0
                P.setValueStencil(row, col, -self.dt * self.du / self.dx ** 2)
                col.field = 1
                row.field = 1
                P.setValueStencil(row, col, -self.dt * self.dv / self.dy ** 2)

        P.assemble()
        if J != P:
            J.assemble()  # matrix-free operator
        return PETSc.Mat.Structure.SAME_NONZERO_PATTERN

    def initial_guess(self) -> VectorPetsc:
        """
        Initial guess
        """
        u = self.dmda.createGlobalVec()
        xa = self.dmda.getVecArray(u)
        for i in range(self.xs, self.xe):
            for j in range(self.ys, self.ye):
                dxx = (i * self.dx)
                dyy = (j * self.dy)
                xa[i, j, 1] = 0
                if dxx >= 1.0 and dxx <= 1.5 and dyy >= 1.0 and dyy <= 1.5:
                    xa[i, j, 1] = (np.sin(4 * np.pi * dxx) ** 2) * (np.sin(4 * np.pi * dyy) ** 2) / 4
                xa[i, j, 0] = 1 - 2 * xa[i, j, 1]
        return VectorPetsc(u)



"""
Vector and application class for the 1D advection problem
"""

import numpy as np
from scipy import sparse as sp
from scipy.sparse.linalg import spsolve
from scipy.sparse import identity

from pymgrit.core.application import Application
from pymgrit.core.vector import Vector


class VectorAdvection1D(Vector):
    """
    Vector class for the 1D advection problem
    """

    def __init__(self, size):
        super().__init__()
        self.size = size
        self.values = np.zeros(size)

    def set_values(self, values):
        self.values = values

    def get_values(self):
        return self.values

    def clone(self):
        tmp = VectorAdvection1D(self.size)
        tmp.set_values(self.get_values())
        return tmp

    def clone_zero(self):
        return VectorAdvection1D(self.size)

    def clone_rand(self):
        tmp = VectorAdvection1D(self.size)
        tmp.set_values(np.random.rand(self.size))
        return tmp

    def __add__(self, other):
        tmp = VectorAdvection1D(self.size)
        tmp.set_values(self.get_values() + other.get_values())
        return tmp

    def __sub__(self, other):
        tmp = VectorAdvection1D(self.size)
        tmp.set_values(self.get_values() - other.get_values())
        return tmp

    def __mul__(self, other):
        tmp = VectorAdvection1D(self.size)
        tmp.set_values(self.get_values() * other)
        return tmp

    def norm(self):
        return np.linalg.norm(self.values)

    def pack(self):
        return self.values

    def unpack(self, values):
        self.values = values


class Advection1D(Application):
    """
    Application class for the advection problem in 1D space,
       u_t + c * u_t = 0,
    subject to periodic boundary conditions in space and
    initial condition u(x, 0) = exp(-x^2).
    """

    def __init__(self, c, x_start, x_end, nx, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.c = c  # (constant) advection speed

        # spatial domain
        self.x_start = x_start  # lower interval bound of spatial domain
        self.x_end = x_end  # upper interval bound of spatial domain
        self.x = np.linspace(self.x_start, self.x_end, nx)
        # periodic boundary conditions
        self.x = self.x[0:-1]
        self.nx = nx - 1
        # spatial grid spacing
        self.dx = self.x[1] - self.x[0]

        self.identity = identity(self.nx, dtype='float', format='csr')

        # set spatial discretization matrix
        self.space_disc = self.compute_matrix()

        # set initial condition
        self.vector_template = VectorAdvection1D(self.nx)
        self.vector_t_start = VectorAdvection1D(self.nx)
        self.initialise()

    def compute_matrix(self):
        """
        Define spatial discretization matrix for advection problem.

        Discretization is first-order upwind in space.
        """

        fac = self.c / self.dx

        diagonal = np.ones(self.nx) * fac
        lower = np.ones(self.nx) * -fac

        matrix = sp.diags(
            diagonals=[diagonal, lower],
            offsets=[0, -1], shape=(self.nx, self.nx),
            format='lil')
        # set periodic entry
        matrix[0, self.nx - 1] = -fac

        return sp.csr_matrix(matrix)

    def initialise(self):
        """
        Set the initial condition of the 1D advection problem
            u(x,0) = exp(-x^2)
        """
        self.vector_t_start.set_values(np.exp(-self.x ** 2))

    def step(self, u_start: VectorAdvection1D, t_start: float, t_stop: float):
        """
        Time integration routine for 1D advection problem:
            Backward Euler

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution for the input time t_stop
        """
        tmp = u_start.get_values()
        tmp = spsolve((t_stop - t_start) * self.space_disc + self.identity, tmp)
        u_stop = VectorAdvection1D(len(tmp))
        u_stop.set_values(tmp)
        return u_stop



"""
AT-MGRIT solver in FAS formulation
"""
import time
import logging
import sys
from operator import itemgetter
from itertools import groupby
import numpy as np

from mpi4py import MPI

from pymgrit.core.mgrit import Mgrit


class AtMgrit(Mgrit):
    def __init__(self, k, *args, **kwargs):
        """
        Constructor

        :param k: Distance of the local coarse grids
        :param args:
        :param kwargs:
        """
        self.k = k
        self.comm_black = None
        self.comm_green = None
        self.local_coarse_grid = None
        self.c_points_per_proc = None
        super().__init__(*args, **kwargs)

    def forward_solve(self, lvl: int) -> None:
        """
        Solves local coarse grid problems on the coarsest level

        :param lvl: AT-MGRIT level
        """

        runtime_fs = time.time()
        if self.lvl_max != 1:
            if self.comm_time_size != 1:
                if self.cpts[lvl].size > 0:
                    data_black = self.comm_black.allgather([
                        self.global_t[lvl][self.cpts[lvl]][-1],
                        self.g[lvl][-1].pack(),
                        self.u[lvl][-1].pack()
                    ])

                    data_green = self.comm_green.bcast(data_black, root=0)

                    data = data_black + data_green

                    tmp_uu = self.problem[lvl].vector_t_start.clone()
                    for i in range(len(data)):
                        if data[i][0] in self.local_coarse_grid:
                            index = np.where(data[i][0] == self.local_coarse_grid)[0][0]
                            if index == 0:
                                tmp_uu.unpack(data[i][2])
                            self.g_coarsest[index].unpack(data[i][1])

                    for i in range(1, len(self.local_coarse_grid)):
                        tmp_uu = self.g_coarsest[i] + self.step[lvl](u_start=tmp_uu,
                                                                     t_start=self.local_coarse_grid[i - 1],
                                                                     t_stop=self.local_coarse_grid[i])
                    if self.comm_time_rank != 0:
                        self.u[lvl][self.index_local_c[lvl][0]] = tmp_uu
                    else:
                        if self.c_points_per_proc[0] == 1:
                            self.u[lvl][self.index_local_c[lvl][0]] = tmp_uu
                        else:
                            self.u[lvl][self.index_local_c[lvl][1]] = tmp_uu

            else:
                tmp_u_arr = [item.clone() for item in self.u[lvl]]
                for point in range(len(self.global_t[lvl])):
                    tmp_u = tmp_u_arr[max(0, point - self.k + 1)]
                    for i in range(max(1, point - self.k + 2), point + 1):
                        tmp_u = self.g[lvl][i] + self.step[lvl](u_start=tmp_u,
                                                                t_start=self.global_t[lvl][i - 1],
                                                                t_stop=self.global_t[lvl][i])
                    self.u[lvl][point] = tmp_u

        logging.debug(f"Forward solve on {self.comm_time_rank} took {time.time() - runtime_fs} s")

    def setup_points_and_comm_info(self, lvl: int) -> None:
        """
        Computes local grid information for level *lvl*.
        Computes which process holds the previous and next point for each lvl and process.

        :param lvl: MGRIT level
        """
        self.global_t.append(np.copy(self.problem[lvl].t))
        points_time = np.size(self.global_t[lvl])
        all_pts = np.array(range(0, points_time))

        # Compute points per process
        # First level: Divide the points evenly
        # Other levels: Depends on the first level
        if lvl == 0:
            block_size_this_lvl, first_i_this_lvl = self.split_points(length=points_time,
                                                                      size=self.comm_time_size,
                                                                      rank=self.comm_time_rank)
            all_pts = all_pts[first_i_this_lvl:first_i_this_lvl + block_size_this_lvl]
            self.int_start = self.global_t[lvl][all_pts[0]]
            self.int_stop = self.global_t[lvl][all_pts[-1]]
        else:
            all_pts = np.where((self.global_t[lvl] >= self.int_start) & (self.global_t[lvl] <= self.int_stop))[0]

        # Compute C- and F-points
        if lvl != self.lvl_max - 1:
            all_cpts = np.where(np.in1d(self.problem[lvl].t, self.problem[lvl + 1].t))[0]
        else:
            all_cpts = np.array(range(0, points_time, self.m[lvl]))
        all_fpts = np.array(list(set(np.array(range(0, points_time))) - set(all_cpts)))
        cpts = np.sort(np.array(list(set(all_pts) - set(all_fpts)), dtype='int'))
        fpts = np.array(list(set(all_pts) - set(cpts)))
        fpts2 = np.array([item for sublist in np.array([np.array(xi, dtype=object) for xi in np.array(
            [list(map(itemgetter(1), g)) for k, g in groupby(enumerate(fpts), lambda x: x[0] - x[1])], dtype=object)],
                                                       dtype=object)[::-1] for item in sublist])

        # Add ghost point if needed, set time interval
        with_ghost_point = False
        if self.comm_time_rank != 0 and all_pts.size > 0:
            with_ghost_point = True
            tmp = np.zeros(len(all_pts) + 1, dtype=int)
            tmp[0] = all_pts[0] - 1
            tmp[1:] = all_pts
            all_pts_with_ghost = tmp
        else:
            all_pts_with_ghost = all_pts

        self.t[lvl] = self.global_t[lvl][all_pts_with_ghost]
        self.cpts.append(cpts)

        # Communication in F-relax
        self.comm_front.append(bool(fpts.size > 0 and fpts[np.argmin(fpts)] - 1 in all_fpts))
        self.comm_back.append(bool(fpts.size > 0 and fpts[np.argmax(fpts)] + 1 in all_fpts))

        # Setup local indices
        self.index_local.append(np.nonzero(all_pts[:, None] == all_pts_with_ghost)[1])
        self.index_local_c.append(np.nonzero(cpts[:, None] == all_pts_with_ghost)[1])
        self.index_local_f.append(np.nonzero(fpts2[:, None] == all_pts_with_ghost)[1])

        # Communication before and after C- and F-relax
        self.first_is_c_point.append(
            bool(all_pts.size > 0 and all_pts[0] in cpts and all_pts[0] != 0 and all_pts[0] - 1 in all_fpts))
        self.first_is_f_point.append(bool(all_pts.size > 0 and all_pts[0] in fpts2 and all_pts[0] - 1 in all_cpts))
        self.last_is_c_point.append(bool(
            all_pts.size > 0 and all_pts[-1] in cpts and all_pts[-1] != points_time - 1 and all_pts[
                -1] + 1 in all_fpts))
        self.last_is_f_point.append(bool(
            all_pts.size > 0 and all_pts[-1] in fpts2 and all_pts[-1] != points_time - 1 and all_pts[
                -1] + 1 in all_cpts))

        # Setup communication info
        split = self.global_t[0][
            np.cumsum(self.split_into(number_points=len(self.global_t[0]), number_processes=self.comm_time_size)) - 1]
        tmp_send_to = -99
        tmp_get_from = -99
        if len(all_pts_with_ghost) > 0:
            if self.t[lvl][-1] != self.global_t[lvl][-1]:
                last_point_next = self.global_t[lvl][np.argwhere(self.global_t[lvl] == self.t[lvl][-1])[0][0] + 1]
                tmp_send_to = np.searchsorted(split, last_point_next)
            if with_ghost_point or self.t[lvl][0] != self.global_t[0][0]:
                tmp_get_from = np.searchsorted(split, self.t[lvl][0])
        self.send_to.append(tmp_send_to)
        self.get_from.append(tmp_get_from)

        # Coarsest level communication
        if lvl == self.lvl_max - 1:

            end_points = np.cumsum(
                self.split_into(number_points=len(self.global_t[0]), number_processes=self.comm_time_size)) - 1
            split = self.global_t[0][end_points]
            self.comm_coarsest_level = np.array([np.min(np.where((item <= split))) for item in self.global_t[-1]])

            values, self.c_points_per_proc = np.unique(self.comm_coarsest_level, return_counts=True)

            if np.any(self.c_points_per_proc[1:] > 1) or self.c_points_per_proc[0] > 2:
                if self.comm_time_size != 1:
                    print('Too many points on proc')
                    sys.exit()

            if self.c_points_per_proc[0] == 2:
                tmp_comm_coarsest_level = self.comm_coarsest_level[1:]
            else:
                tmp_comm_coarsest_level = self.comm_coarsest_level

            # New communicators for MGRIT-DD
            if self.comm_time_rank in tmp_comm_coarsest_level:
                idx = np.where(self.comm_time_rank == tmp_comm_coarsest_level)[0][0]
                color_black = idx // self.k
                if idx % self.k == self.k - 1:
                    color_green = color_black + 1
                else:
                    color_green = color_black
                if self.comm_time_rank != 0:
                    self.local_coarse_grid = np.array(
                        self.global_t[lvl][max(0, self.cpts[lvl][0] - self.k + 1):self.cpts[lvl][0] + 1])
                else:
                    if self.c_points_per_proc[0] == 1:
                        self.local_coarse_grid = np.array(
                            self.global_t[lvl][max(0, self.cpts[lvl][0] - self.k + 1):self.cpts[lvl][0] + 1])
                    else:
                        self.local_coarse_grid = np.array(
                            self.global_t[lvl][max(0, self.cpts[lvl][0] - self.k + 1):self.cpts[lvl][1] + 1])

            else:
                color_black = MPI.UNDEFINED
                color_green = MPI.UNDEFINED

            self.comm_black = self.comm_time.Split(color=color_black, key=self.comm_time_rank)
            self.comm_green = self.comm_time.Split(color=color_green, key=self.comm_time_rank)

    def create_coarsest_level(self) -> None:
        """
        Creates vectors u and g for coarsest level
        """
        if self.local_coarse_grid is not None:
            self.g_coarsest = [self.problem[-1].vector_template.clone_zero() for _ in
                               range(len(self.local_coarse_grid))]
        self.u_coarsest = [self.problem[-1].vector_template.clone_zero() for _ in range(len(self.global_t[-1]))]
        self.u_coarsest[0] = self.problem[-1].vector_t_start.clone()

    def ouput_run_information(self) -> None:
        """
        Outputs information of AT-MGRIT run.
        """
        msg = ['Run parameter overview',
               '  ' + '{0: <25}'.format(f'time interval') + ' : ' + '[' + str(self.problem[0].t[0]) + ', ' + str(
                   self.problem[0].t[-1]) + ']',
               '  ' + '{0: <25}'.format(f'number of time points ') + ' : ' + str(len(self.problem[0].t)),
               '  ' + '{0: <25}'.format(f'max dt ') + ' : ' + str(
                   np.max(self.problem[0].t[1:] - self.problem[0].t[:-1])),
               '  ' + '{0: <25}'.format(f'number of levels') + ' : ' + str(self.lvl_max),
               '  ' + '{0: <25}'.format(f'coarsening factors') + ' : ' + str(self.m[:-1]),
               '  ' + '{0: <25}'.format(f'relaxation weight') + ' : ' + str(self.weight_c),
               '  ' + '{0: <25}'.format(f'cf_iter') + ' : ' + str(self.cf_iter),
               '  ' + '{0: <25}'.format(f'nested iteration') + ' : ' + str(self.nes_it),
               '  ' + '{0: <25}'.format(f'cycle type') + ' : ' + str(self.cycle_type),
               '  ' + '{0: <25}'.format(f'stopping tolerance') + ' : ' + str(self.tol),
               '  ' + '{0: <25}'.format(f'time communicator size') + ' : ' + str(self.comm_time_size),
               '  ' + '{0: <25}'.format(f'space communicator size') + ' : ' + str(self.comm_space_size),
               '  ' + '{0: <25}'.format(f'distance') + ' : ' + str(self.k)]
        self.log_info(message='\n'.join(msg))



"""
Application class for the 2D heat equation
"""

import numpy as np

from mpi4py import MPI

from pymgrit.core.grid_transfer import GridTransfer
from pymgrit.petsc.vector_petsc import VectorPetsc
from pymgrit.core.application import Application

try:
    from petsc4py import PETSc
except ImportError as e:
    import sys

    sys.exit("This example requires petsc4py.")


class HeatPetsc(Application):
    """
    2D heat equation application with Dirichlet BCs in [0,1]x[0,1]
    """

    def __init__(self, dmda: PETSc.DMDA, comm_x: MPI.Comm, freq: int, a: float, rtol: float = 1e-10,
                 atol: float = 1e-10, max_it: int = 100, *args, **kwargs) -> None:
        """
        Constructor
        :param dmda: PETSc DMDA grid
        :param comm_x: space communicator
        :param freq: frequency
        :param a: diffusion coefficient
        :param rtol: spatial solver tolerance
        :param atol: spatial solver tolerance
        :param max_it: spatial solver max iter
        :param args:
        :param kwargs:
        """
        super().__init__(*args, **kwargs)
        self.dmda = dmda
        self.mx, self.my = self.dmda.sizes
        self.dx = 1.0 / (self.mx - 1)
        self.dy = 1.0 / (self.my - 1)
        (self.xs, self.xe), (self.ys, self.ye) = self.dmda.ranges
        self.freq = freq
        self.a = a
        self.space_disc = self.get_a()
        self.id = self.get_id()

        self.vector_template = VectorPetsc(self.dmda.createGlobalVec())
        self.vector_t_start = VectorPetsc(self.u_exact(0).get_values())

        # setup solver
        self.ksp = PETSc.KSP()
        self.ksp.create(comm=comm_x)
        self.ksp.setType('gmres')
        pc = self.ksp.getPC()
        pc.setType('none')
        self.ksp.setFromOptions()
        self.ksp.setTolerances(rtol=rtol, atol=atol, max_it=max_it)

        self.dt = self.t[1] - self.t[0]
        self.ksp.setOperators(self.dt * self.space_disc + self.id)

    def step(self, u_start: VectorPetsc, t_start: float, t_stop: float) -> VectorPetsc:
        """
        Time integration routine for 2D heat equation example problem

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        result = self.dmda.createGlobalVec()
        if not np.isclose(self.dt, (t_stop - t_start)):
            self.ksp.setOperators((t_stop - t_start) * self.space_disc + self.id)
            self.dt = (t_stop - t_start)
        self.ksp.solve(self.compute_rhs(u_start=u_start, t_start=t_start, t_stop=t_stop), result)

        return VectorPetsc(result)

    def get_a(self) -> PETSc.Mat:
        """
        Define spatial discretization matrix for 2D heat equation

        Second-order central finite differences with matrix stencil
            [          -f_y          ]
            [-f_x  2(f_x + f_y)  -f_x]
            [          -f_y          ]
        with f_x = (a / dx^2) and f_y = (a / dy^2)
        :return: spatial discretization
        """
        A = self.dmda.createMatrix()
        A.setType('aij')
        A.setFromOptions()
        A.setPreallocationNNZ((5, 5))
        A.setUp()

        fx = self.a / self.dx ** 2
        fy = self.a / self.dy ** 2

        A.zeroEntries()
        row = PETSc.Mat.Stencil()
        col = PETSc.Mat.Stencil()
        for j in range(self.ys, self.ye):
            for i in range(self.xs, self.xe):
                row.index = (i, j)
                row.field = 0
                if i == 0 or j == 0 or i == self.mx - 1 or j == self.my - 1:
                    A.setValueStencil(row, row, 0.0)
                else:
                    diag = 2 * (fx + fy)
                    for index, value in [
                        ((i, j - 1), -fx),
                        ((i - 1, j), -fy),
                        ((i, j), diag),
                        ((i + 1, j), -fy),
                        ((i, j + 1), -fx),
                    ]:
                        col.index = index
                        col.field = 0
                        A.setValueStencil(row, col, value)
        A.assemble()
        return A

    def get_id(self) -> PETSc.Mat:
        """
        Define identity matrix
        :return: identity matrix
        """
        Id = self.dmda.createMatrix()
        Id.setType('aij')
        Id.setFromOptions()
        Id.setPreallocationNNZ((1, 1))
        Id.setUp()

        Id.zeroEntries()
        row = PETSc.Mat.Stencil()
        (xs, xe), (ys, ye) = self.dmda.ranges
        for j in range(ys, ye):
            for i in range(xs, xe):
                row.index = (i, j)
                row.field = 0
                Id.setValueStencil(row, row, 1.0)
        Id.assemble()
        return Id

    def compute_rhs(self, u_start: VectorPetsc, t_start: float, t_stop: float) -> PETSc.Vec:
        """
        Right-hand side of spatial system

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: right-hand side of spatial system at each time step in case of implicit time integration
        """
        b = self.dmda.createGlobalVec()
        u = u_start.get_values()

        ba = self.dmda.getVecArray(b)
        ua = self.dmda.getVecArray(u)

        ba[self.xs:self.xe, self.ys:self.ye] = ua[self.xs:self.xe, self.ys:self.ye] + (
                t_stop - t_start) * self.rhs(t_stop)
        return b

    def rhs(self, t_stop: float) -> np.ndarray:
        """
        Right hand side
        :param t_stop: time point
        :return: right hand side
        """
        xv, yv = np.meshgrid(range(self.xs, self.xe), range(self.ys, self.ye), indexing='ij')
        res = -np.sin(np.pi * self.freq * xv * self.dx) * \
              np.sin(np.pi * self.freq * yv * self.dy) * \
              (np.sin(t_stop) - self.a * 2.0 * (np.pi * self.freq) ** 2 * np.cos(t_stop))
        return res

    def u_exact(self, t: float) -> VectorPetsc:
        """
        Exact solution
        :param t: time point
        :return: exact solution at point t
        """
        u = self.dmda.createGlobalVec()
        xa = self.dmda.getVecArray(u)
        for i in range(self.xs, self.xe):
            for j in range(self.ys, self.ye):
                xa[i, j] = np.sin(np.pi * self.freq * i * self.dx) * \
                           np.sin(np.pi * self.freq * j * self.dy) * np.cos(t)

        return VectorPetsc(u)


class GridTransferPetsc(GridTransfer):
    """
    Grid Transfer class between PETSc DMDA grids
    """

    def __init__(self, fine_prob: PETSc.DMDA, coarse_prob: PETSc.DMDA) -> None:
        """
        Constructor
        :param fine_prob:
        :param coarse_prob:
        """
        super().__init__()

        self.coarse_prob = coarse_prob
        self.fine_prob = fine_prob
        self.interp, _ = self.coarse_prob.createInterpolation(fine_prob)
        self.inject = self.coarse_prob.createInjection(fine_prob)

    def restriction(self, u: VectorPetsc) -> VectorPetsc:
        """
        Restriction
        :param u: fine approximation
        :return: coarse approximation
        """
        u_coarse = self.coarse_prob.createGlobalVec()
        self.inject.mult(u.get_values(), u_coarse)
        return VectorPetsc(u_coarse)

    def interpolation(self, u: VectorPetsc) -> VectorPetsc:
        """
        Interpolation
        :param u: coarse approximation
        :return: fine approximation
        """
        u_fine = self.fine_prob.createGlobalVec()
        self.interp.mult(u.get_values(), u_fine)
        return VectorPetsc(u_fine)



"""
Vector class for Firedrake applications
"""

try:
    from firedrake import norm, Function
except ImportError as e:
    import sys

    sys.exit("This example requires firedrake. See https://pymgrit.github.io/pymgrit/coupling/firedrake.html")

import numpy as np

from pymgrit.core.vector import Vector


class VectorFiredrake(Vector):
    """
    Vector class for Firedrake Function object
    """

    def __init__(self, values: Function):
        """
        Constructor.
        """

        super().__init__()
        if isinstance(values, Function):
            self.values = values.copy(deepcopy=True)
        else:
            raise Exception('Wrong datatype')

    def set_values(self, values):
        """
        Set vector data

        :param values: values for vector object
        """
        self.values = values

    def get_values(self):
        """
        Get vector data

        :return: values of vector object
        """
        return self.values

    def clone(self):
        """
        Initialize vector object with copied values

        :rtype: vector object with zero values
        """
        return VectorFiredrake(self.values)

    def clone_zero(self):
        """
        Initialize vector object with zeros

        :rtype: vector object with zero values
        """
        tmp = VectorFiredrake(self.values)
        tmp = tmp * 0
        return tmp

    def clone_rand(self):
        """
        Initialize vector object with random values

        :rtype: vector object with random values
        """
        tmp = VectorFiredrake(self.values)
        tmp_values = tmp.get_values()
        tmp_values.dat.data[:] = np.random.rand(*tmp_values.dat.data[:].shape)
        tmp.set_values(tmp_values)
        return tmp

    def __add__(self, other):
        """
        Addition of two vector objects (self and other)

        :param other: vector object to be added to self
        :return: sum of vector object self and input object other
        """
        tmp = VectorFiredrake(self.values)
        tmp_value = tmp.get_values()
        tmp_value += other.get_values()
        tmp.set_values(tmp_value)
        return tmp

    def __sub__(self, other):
        """
        Subtraction of two vector objects (self and other)

        :param other: vector object to be subtracted from self
        :return: difference of vector object self and input object other
        """
        tmp = VectorFiredrake(self.values)
        tmp_value = tmp.get_values()
        tmp_value -= other.get_values()
        tmp.set_values(tmp_value)
        return tmp

    def __mul__(self, other):
        """
        Multiplication of a vector object and a float (self and other)

        :param other: object to be multiplied with self
        :return: difference of vector object self and input object other
        """
        tmp = VectorFiredrake(self.values)
        tmp_value = tmp.get_values()
        tmp_value *= other
        tmp.set_values(tmp_value)
        return tmp

    def norm(self):
        """
        Norm of a vector object

        :return: 2-norm of vector object
        """
        return norm(self.values)

    def unpack(self, values):
        """
        Unpack and set data

        :param values: values for vector object
        """
        self.values.dat.data[:] = values

    def pack(self):
        """
        Pack data

        :return: values of vector object
        """
        return self.values.dat.data[:]



"""
Abstract application class for user-defined application classes.
Every user-defined application class must inherit from this class.

Required attributes:
  - vector_template
  - vector_t_start
Required functions:
  - step
"""
from abc import ABCMeta, abstractmethod
import numpy as np

from pymgrit.core.vector import Vector


class MetaApplication(ABCMeta):
    """
    MetaClass for the application class.
    Checks whether required attributes are set.
    """
    required_attributes = []

    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        for attr_name in obj.required_attributes:
            if not hasattr(obj, attr_name):
                raise ValueError('required attribute (%s) not set' % attr_name)
        return obj


class Application(object, metaclass=MetaApplication):
    """
    Abstract application class for user-defined application classes.
    Every user-defined application class must inherit from this class.

    Required attributes:
      - vector_template
      - vector_t_start
    Required functions:
      - step
    """
    required_attributes = ['vector_template', 'vector_t_start']

    def __init__(self, t_start: float = None, t_stop: float = None, nt: int = None,
                 t_interval: np.ndarray = None) -> None:
        """
        Initialize time information

        :param t_start: Start time (left bound of time interval)
        :param t_stop: End time (right bound of time interval)
        :param nt: Number of time points
        """

        if t_interval is None:
            if t_start is None or t_stop is None or nt is None:
                raise Exception('Specify an interval by t_start, t_stop and nt or by t_interval')
            self.t_start = t_start
            self.t_end = t_stop
            self.nt = nt
            self.t = np.linspace(self.t_start, self.t_end, nt)
        else:
            if not isinstance(t_interval, np.ndarray):
                raise Exception('t_interval has the wrong type. Should be a numpy array')
            self.t_start = t_interval[0]
            self.t_end = t_interval[-1]
            self.nt = len(t_interval)
            self.t = t_interval

    @property
    def vector_template(self) -> Vector:
        """
        Getter for attribute vector_template
        """
        return self._vector_template

    @vector_template.setter
    def vector_template(self, value: Vector) -> None:
        """
        Setter for attribute vector_template
        """
        self._vector_template = value

    @property
    def vector_t_start(self) -> Vector:
        """
        Getter for attribute vector_t_start
        """
        return self._vector_t_start

    @vector_t_start.setter
    def vector_t_start(self, value: Vector) -> None:
        """
        Setter for attribute vector_t_start
        """
        self._vector_t_start = value

    @abstractmethod
    def step(self, u_start: object, t_start: float, t_stop: float) -> object:
        """
        Time integration routine for the application

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """



"""
Application class for the 2D heat equation
"""

try:
    from firedrake import FunctionSpace, Constant, TestFunction, TrialFunction, Function, FacetNormal, inner, dx, grad
    from firedrake import outer, LinearVariationalProblem, NonlinearVariationalSolver, dS, exp, SpatialCoordinate, avg
except ImportError as e:
    import sys

    sys.exit("This example requires firedrake. See https://pymgrit.github.io/pymgrit/coupling/firedrake.html")

from mpi4py import MPI

from pymgrit.core.application import Application
from pymgrit.firedrake.vector_firedrake import VectorFiredrake


class Diffusion2D(Application):
    """
    Application class containing the description of the diffusion problem.

    The spatial discretisation is P1 DG (piecewise linear discontinous
    elements) and uses an interior penalty method which penalises jumps
    at element interfaces.
    """

    def __init__(self, mesh: object, kappa: float, comm_space: MPI.Comm, mu: float = 5., *args, **kwargs):
        """
        Constructor

        :param mesh: spatial domain
        :param kappa: diffusion coefficient
        :param mu: penalty weighting function
        """
        super().__init__(*args, **kwargs)

        # Spatial domain and function space
        self.mesh = mesh
        self.function_space = FunctionSpace(self.mesh, "DG", 1)
        self.comm_space = comm_space

        # Placeholder for time step - will be updated in the update method
        self.dt = Constant(0.)

        # Things we need for the form
        gamma = TestFunction(self.function_space)
        phi = TrialFunction(self.function_space)
        self.f = Function(self.function_space)
        n = FacetNormal(mesh)

        # Set up the rhs and bilinear form of the equation
        a = (inner(gamma, phi) * dx
             + self.dt * (
                     inner(grad(gamma), grad(phi) * kappa) * dx
                     - inner(2 * avg(outer(phi, n)), avg(grad(gamma) * kappa)) * dS
                     - inner(avg(grad(phi) * kappa), 2 * avg(outer(gamma, n))) * dS
                     + mu * inner(2 * avg(outer(phi, n)), 2 * avg(outer(gamma, n) * kappa)) * dS
             )
             )
        rhs = inner(gamma, self.f) * dx

        # Function to hold the solution
        self.soln = Function(self.function_space)

        # Setup problem and solver
        prob = LinearVariationalProblem(a, rhs, self.soln)
        self.solver = NonlinearVariationalSolver(prob)

        # Set the data structure for any user-defined time point
        self.vector_template = VectorFiredrake(self.soln)

        # Set initial condition:
        # Setting up a Gaussian blob in the centre of the domain.
        x = SpatialCoordinate(self.mesh)
        initial_tracer = exp(-((x[0] - 5) ** 2 + (x[1] - 5) ** 2))
        tmp = Function(self.function_space)
        tmp.interpolate(initial_tracer)
        self.vector_t_start = VectorFiredrake(tmp)

    def step(self, u_start: VectorFiredrake, t_start: float, t_stop: float) -> VectorFiredrake:
        """
        Time integration routine for 2D diffusion problem:
            Backward Euler

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        # Time-step size
        self.dt.assign(t_stop - t_start)

        self.f.assign(u_start.get_values())

        # Take Backward Euler step
        self.solver.solve()

        return VectorFiredrake(self.soln)



"""
Application class for the model 'im_3_kW' of an induction machine
https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines
"""

from typing import Dict
import os
import subprocess
from subprocess import PIPE
import tempfile

import numpy as np

from pymgrit.core.application import Application
from pymgrit.induction_machine.vector_machine import VectorMachine
from pymgrit.induction_machine.helper import is_numeric, pre_file, get_values_from, getdp_read_resolution, \
    set_resolution, get_preresolution


class InductionMachine(Application):
    """
    Simulating an induction machine using the model 'im_3_kW' from
    https://gitlab.onelab.info/doc/models/-/wikis/Electric-machines
    """

    def __init__(self, grid: str, path_im3kw: str, path_getdp: str, imposed_speed: int = 1, nb_trelax: int = 2,
                 analysis_type: int = 1, nb_max_iter: int = 60, relaxation_factor: float = 0.5,
                 stop_criterion: float = 1e-6, nonlinear: bool = False, pwm: bool = False, pro_file: str = 'im_3kW.pro',
                 verbose: bool = False, steps_per_solve: int = 1, *args, **kwargs):
        """
        Constructor

        :param nonlinear: Nonlinear or linear model
        :param pwm: pwm or sin rhs
        :param grid: mesh
        :param path_im3kw: path to im_3kW data
        :param path_getdp: path to getdp
        :param imposed_speed: imposed speed
        """
        super().__init__(*args, **kwargs)

        self.pro_path = path_im3kw + pro_file
        if not os.path.isfile(self.pro_path):
            raise Exception('Found no valid .pro file in', self.pro_path)

        self.getdp_path = path_getdp
        if not os.path.isfile(self.getdp_path):
            raise Exception('Getdp not found (http://getdp.info/)')

        self.nl = int(nonlinear)
        self.pwm = int(pwm)
        self.mesh = grid + '.msh'
        self.pre = grid + '.pre'
        self.further_unknowns_front = 8
        self.further_unknowns_back = 15
        self.steps_per_solve = steps_per_solve

        cor_to_un, un_to_cor, boundary = pre_file(path_im3kw + self.pre)

        self.nx = len(un_to_cor) + self.further_unknowns_front + self.further_unknowns_back

        self.gopt = {'Verbose': int(verbose), 'TimeStep': (self.t[1] - self.t[0]) / self.steps_per_solve,
                     'Executable': self.getdp_path, 'PreProcessing': '#1'}
        self.fopt = ['Flag_AnalysisType', analysis_type, 'Flag_NL', self.nl, 'Flag_ImposedSpeed', imposed_speed,
                     'Nb_max_iter', nb_max_iter, 'relaxation_factor', relaxation_factor, 'stop_criterion',
                     stop_criterion, 'NbTrelax', nb_trelax, 'Flag_PWM', self.pwm]

        version_test = subprocess.run([self.gopt['Executable'], '--version'], stdout=PIPE, stderr=PIPE)
        if version_test.returncode:
            raise Exception('getdp not found.')

        self.vector_template = VectorMachine(u_front_size=self.further_unknowns_front,
                                             u_back_size=self.further_unknowns_back,
                                             u_middle_size=len(un_to_cor))
        self.vector_t_start = VectorMachine(u_front_size=self.further_unknowns_front,
                                            u_back_size=self.further_unknowns_back,
                                            u_middle_size=len(un_to_cor))

    def step(self, u_start: VectorMachine, t_start: float, t_stop: float) -> VectorMachine:
        """
        Time integration routine

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        soli = self.run_getdp(u_start=u_start.get_values(), t_start=t_start, t_stop=t_stop)
        ret = VectorMachine(u_front_size=u_start.u_front_size,
                            u_back_size=u_start.u_back_size, u_middle_size=u_start.u_middle_size)

        ret.set_values(values=soli['y'][-1], jl=soli['jl'][-1], ia=soli['ia'][-1], ib=soli['ib'][-1], ic=soli['ic'][-1],
                       ua=soli['ua'][-1], ub=soli['ub'][-1], uc=soli['uc'][-1], tr=soli['tr'][-1])
        return ret

    def run_getdp(self, u_start: np.ndarray, t_start: float, t_stop: float) -> Dict:
        """
        Calls getdp

        :param u_start: approximate solution for the input time t_start
        :param t_start: time associated with the input approximate solution u_start
        :param t_stop: time to evolve the input approximate solution to
        :return: approximate solution at input time t_stop
        """
        if np.max(np.isnan(u_start)):
            raise Exception('Approximation contains nan')

        fdir, file = os.path.split(self.pro_path)
        fname, fext = os.path.splitext(file)

        funargstr = ''
        for i in range(0, len(self.fopt), 2):
            if is_numeric(self.fopt[i + 1]):
                funargstr = funargstr + ' -setnumber ' + str(self.fopt[i]) + ' ' + str(self.fopt[i + 1])
            else:
                funargstr = funargstr + ' -setstring ' + str(self.fopt[i]) + ' ' + self.fopt[i + 1]
        funargstr = funargstr[1:]

        mshfile = os.path.join(fdir, self.mesh)

        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_name = os.path.join(tmpdir, fname)
            resdir = os.path.join(tmpdir, 'res')
            prefile = os.path.join(tmpdir, fname + '.pre')
            resfile = os.path.join(tmpdir, fname + '.res')
            joule_file = os.path.join(tmpdir, 'resJL.dat')
            ua_file = os.path.join(tmpdir, 'resUa.dat')
            ub_file = os.path.join(tmpdir, 'resUb.dat')
            uc_file = os.path.join(tmpdir, 'resUc.dat')
            ia_file = os.path.join(tmpdir, 'resIa.dat')
            ib_file = os.path.join(tmpdir, 'resIb.dat')
            ic_file = os.path.join(tmpdir, 'resIc.dat')
            tr_file = os.path.join(tmpdir, 'resTr.dat')

            # Preprocessing
            exe_string = [self.gopt['Executable'],
                          self.pro_path,
                          '-pre "' + self.gopt['PreProcessing'] + '"',
                          '-msh', mshfile,
                          '-name', tmp_name,
                          '-res', resfile,
                          '-setnumber timemax', str(t_stop),
                          '-setnumber dtime', str(self.gopt['TimeStep']),
                          '-setstring ResDir', resdir,
                          funargstr]

            if self.gopt['Verbose'] == 1:
                status = subprocess.run(' '.join(exe_string), shell=True)
            else:
                status = subprocess.run(' '.join(exe_string), shell=True, stdout=PIPE, stderr=PIPE)

            if status.returncode:
                raise Exception('preprocessing failed')

            num_dofs = np.size(u_start)
            num_pres = get_preresolution(file=prefile)

            if num_dofs != np.sum(num_pres):
                raise Exception(
                    'u_start has wrong size: ' + str(num_dofs) + ' instead of ' + str(num_pres) + ': ' + str(prefile))

            # Create initial data
            set_resolution(file=resfile, t_start=t_start, u_start=u_start, num_dofs=num_dofs)

            # Compute solution
            exe_string = [self.gopt['Executable'],
                          self.pro_path,
                          '-restart',
                          '-msh', mshfile,
                          '-name', tmp_name,
                          '-res', resfile,
                          '-setnumber timemax', str(t_stop),
                          '-setnumber dtime', str(self.gopt['TimeStep']),
                          '-setstring ResDir', resdir,
                          funargstr]

            if self.gopt['Verbose'] == 1:
                status = subprocess.run(' '.join(exe_string), shell=True)
            else:
                status = subprocess.run(' '.join(exe_string), shell=True, stdout=PIPE, stderr=PIPE)
            if status.returncode:
                raise Exception('getdp solving failed')

            # Read results
            t, y = getdp_read_resolution(file=resfile, num_dofs=num_dofs)
            jl = get_values_from(file=joule_file)
            ia = get_values_from(file=ia_file)
            ib = get_values_from(file=ib_file)
            ic = get_values_from(file=ic_file)
            ua = get_values_from(file=ua_file)
            ub = get_values_from(file=ub_file)
            uc = get_values_from(file=uc_file)
            tr = get_values_from(file=tr_file)

        return {'x': t, 'y': y, 'jl': jl, 'ia': ia, 'ib': ib, 'ic': ic, 'ua': ua, 'ub': ub, 'uc': uc, 'tr': tr}
